from .types import PointMap
from contextlib import contextmanager
from pyadjoint.enlisting import Enlist
from pyadjoint import annotate_tape, get_working_tape, stop_annotating
from pyadjoint_utils import Tape, push_tape
from pyadjoint_utils.fenics_adjoint import assemble
from pyadjoint_utils.tape_block import record_tape_block, TapeBlock
from pyadjoint_utils import ReducedFunctionNumPy
from pyadjoint_utils.identity import JacobianIdentity, make_jacobian_identities
from dolfin.fem.formmanipulations import derivative
from ufl import replace, dot, inner, grad, Form, Coefficient
from ufl.algorithms import expand_derivatives
from ufl.core.expr import Expr
from ufl.log import error
from crikit.fe import contraction
from ..covering import get_composite_cr
from .ufl import UFLExprSpace, UFLFunctionSpace
from .space_builders import DirectSum
from ..cr.quadrature import (
    get_quadrature_params,
    make_quadrature_spaces,
    make_quadrature_space,
)
from ..cr.quadrature import estimate_total_polynomial_degree as est_degree
from .map_builders import CompositePointMap
import numpy as np
import jax.numpy as jnp
import jax
import ufl

from typing import Tuple, Union, Optional
import backend
from fenics_adjoint.compat import compat

compat = compat(backend)


def form_get_expr_space(form, exprs):
    terms = Enlist(exprs)
    domain = form.ufl_domain()
    spaces = tuple(UFLExprSpace(t, ufl_domains=(domain,)) for t in terms)
    space = terms.delist(spaces)
    if not isinstance(space, UFLExprSpace):
        space = DirectSum(*space)
    return space


def _assemble_with_cr(
    form, cr, arg, out_terms, quad_params=None, force_explicit=False, **kwargs
):
    """Substitute the output of cr applied to arg (tuple of UFL Exprs) in the form.

    If the cr does not handle UFL Expr inputs, convert to a quadrature element."""
    annotate = annotate_tape(kwargs)
    b_kwargs = AssembleWithCRBlock.pop_kwargs(kwargs)
    b_kwargs.update(kwargs)

    with push_tape() as assemble_tape:
        # First, the CR is converted to a ufl CR if necessary.
        explicit = force_explicit or not cr.source.is_point(arg)
        if not explicit:
            with stop_annotating():
                newexprs = cr(arg)
            term_space = form_get_expr_space(form, out_terms)
            explicit = not term_space.is_point(newexprs)
        if explicit:
            form, ufl_cr, quad_params = get_ufl_composite_cr(
                form, cr, arg, out_terms, quad_params=quad_params
            )
        else:
            ufl_cr = cr

        # Then, the CR is run on a separate tape (to facilitate faster Jacobian calculations).
        with record_tape_block(name="FullCRTapeBlock") as cr_tape_full:
            if explicit:
                # This assumes the cr consists of three point maps: input Covering, inner CR, and output Covering.
                point_maps = ufl_cr.point_maps()
                if len(point_maps) != 3:
                    raise ValueError(
                        "Expected 3 point maps in composite cr (got %d)"
                        % len(point_maps)
                    )
                if point_maps[1] != cr:
                    raise ValueError(
                        "Expected second point map in composite cr to be the original cr (%s != %s)"
                        % (point_maps[1], cr)
                    )

                inner_inputs = point_maps[0](arg)
                with record_tape_block(name="InnerCRTapeBlock") as cr_tape_inner:
                    inner_outputs = point_maps[1](inner_inputs)
                cr_outputs = point_maps[2](inner_outputs)

                cr_tape_info = {
                    "inner_inputs": Enlist(inner_inputs),
                    "cr_tape_inner": cr_tape_inner,
                    "inner_outputs": Enlist(inner_outputs),
                    "cr_outputs": Enlist(cr_outputs),
                }
            else:
                cr_tape_info = {}
                cr_outputs = ufl_cr(arg)

        # Finally, the outputs of the CR are put into the form, and the form is assembled.
        out_map = {}
        for expr, term in zip(Enlist(cr_outputs), Enlist(out_terms)):
            out_map[term] = expr
        newform = replace(form, out_map)
        a = assemble(newform, **kwargs)

    if annotate:
        working_tape = get_working_tape()
        block = AssembleWithCRBlock(
            assemble_tape,
            form,
            cr,
            arg,
            out_terms,
            quad_params=quad_params,
            cr_tape_full=cr_tape_full,
            explicit=explicit,
            **cr_tape_info,
            **b_kwargs,
        )
        working_tape.add_block(block)
    return a, form, ufl_cr


def assemble_with_cr(
    form: Form,
    cr: PointMap,
    arg: Union[Expr, Tuple[Expr]],
    out_terms: Union[Coefficient, Tuple[Coefficient]],
    quad_params: Optional[dict] = None,
    force_explicit: Optional[bool] = False,
    return_all: Optional[bool] = False,
    **kwargs,
):
    """Substitute the output of ``cr`` applied to ``arg`` in the form and then assemble the form.

    Given a UFL form that contains ``out_terms``, this function calculates
    ``cr_out = cr(arg)``, replaces each term in ``out_terms`` with the
    corresponding term in ``cr_out``, and assembles the resulting form.

    If ``cr`` can't take ``arg`` directly as input, or if it doesn't output
    Functions, or if ``force_explicit=True``, then
    :func:`~crikit.covering.covering.get_composite_cr` used to to map ``arg`` to
    the CR's input space and to map the CR's output to a UFL function space.

    Args:
        form (~ufl.classes.Form): the form
        cr (PointMap): a point map whose input is compatible with ``arg`` and
            whose output is compatible with ``out_terms``.
        arg (~ufl.classes.Expr or tuple[~ufl.classes.Expr]): input expressions to
            the CR
        out_terms (~ufl.classes.Coefficient or tuple[~ufl.classes.Coefficient]):
            terms in the form that will be replaced by the output of the CR
        quad_params (dict, optional): parameters for the quadrature space
        force_explicit (bool, optional): pass ``True`` to force
            ``get_composite_cr`` to be called.
        return_all (bool, optional): Set to ``True`` to have the unassemled form
            and composite CR returned.

        **kwargs: passed through to the ``assemble`` function of the backend

    Returns:
        float or ~ufl.classes.Coefficient: the output of ``pyadjoint_utils.fenics_adjoint.assemble``

        If ``return_all`` is true, additionally returns:

            * :class:`~ufl.classes.Form`: the form with the output of the CR inserted
            * :class:`~crikit.cr.types.PointMap`: the output of :func:`~crikit.covering.covering.get_composite_cr`

    """
    a, form, ufl_cr = _assemble_with_cr(
        form, cr, arg, out_terms, quad_params, force_explicit, **kwargs
    )

    if return_all:
        return a, form, ufl_cr
    return a


def get_ufl_composite_cr(form, cr, arg_terms, out_terms, quad_params=None):
    arg_terms = Enlist(arg_terms)
    out_terms = Enlist(out_terms)
    degree = get_cr_form_degree(form, cr, arg_terms, out_terms)
    quad_params = get_quadrature_params(quad_params, degree)

    # This assumes the form only has one domain, which is also an assumption
    # dolfin makes, so it should be good.
    domain = form.ufl_domain()

    # Create quadrature spaces for output.
    quad_spaces, quad_params = make_quadrature_spaces(
        tuple(UFLExprSpace(o, ufl_domains=(domain,)) for o in out_terms),
        quad_params=quad_params,
        domain=domain,
    )

    # Build a composite CR to work with arg_terms as input and out_funcs as output.
    input_space = arg_terms.delist(
        DirectSum(*tuple(UFLExprSpace(a, ufl_domains=(domain,)) for a in arg_terms))
    )
    output_space = out_terms.delist(
        DirectSum(*tuple(UFLExprSpace(o, ufl_domains=(domain,)) for o in out_terms))
    )
    # output_space = out_terms.delist(DirectSum(tuple(UFLFunctionSpace(Q) for Q in quad_spaces)))

    composite_cr = get_composite_cr(
        input_space, cr, output_space, domain=domain, quad_params=quad_params
    )

    # Add the quadrature info to the form by updating the metadata of each integral.
    new_metadata = [ig.metadata().copy() for ig in form.integrals()]
    for m in new_metadata:
        m.update(quad_params)
    new_integrals = [
        ig.reconstruct(metadata=m) for ig, m in zip(form.integrals(), new_metadata)
    ]
    new_form = Form(new_integrals)

    return new_form, composite_cr, quad_params


def get_cr_form_degree(form, cr, arg_terms, out_terms):
    # Get degree of all arguments to the CR.
    arg_degrees = tuple(est_degree(a) for a in arg_terms)
    max_arg_degree = max(arg_degrees)
    form_degree = est_degree(form, default_degree=max_arg_degree)

    # Get the degrees of the outputs of the CR.
    cr_degrees = cr.est_degree(*arg_degrees)
    if cr_degrees is not None:
        cr_degrees = Enlist(cr_degrees)
        assert len(cr_degrees) == len(out_terms)

        # Get the degree of the form by telling UFL the degree of each output of the CR.
        cr_degree_map = dict(
            (coeff, degree) for coeff, degree in zip(out_terms, cr_degrees)
        )
        d = est_degree(
            form, default_degree=max_arg_degree, coefficient_replace_map=cr_degree_map
        )
        form_degree = max(form_degree, d)
    degree = max(max_arg_degree, form_degree)
    return degree


class AssembleWithCRBlock(TapeBlock):
    """This block represents both an AssemblyBlock (for the assemble call) and a
    TapeBlock (for the CR call).

    Almost all calls are passed through to the original tape, which consists of
    calling a CR and running assemble(). This block optimizes the tlm matrix
    computation by avoiding the Covering projection calls and inserting the CR's
    Jacobian directly into the form for assemble().

    Steps:
        1. Get CR Jacobian.
        2. Insert it into form by replacing out_terms with Jacobian contracted with arg_terms.
        3. Replace coefficients in arg_terms with the appropriate UFL arguments.
    """

    pop_kwargs_keys = ["name", "tlm_mat_skip_covering"]

    def __init__(self, tape, form, cr, arg_terms, out_terms, **kwargs):
        super().__init__(tape)
        self._form = form
        self._cr = cr
        self._arg_terms = Enlist(arg_terms)
        self._out_terms = Enlist(out_terms)

        # There are three tapes to keep track of.
        # 1. self.tape: this is the tape that tracks everything done in this block.
        # 2. self.cr_tape_full: this is the tape that records the CR call, including the Covering stuff.
        # 3. self.cr_tape_inner: this is the tape that records the CR call, not including the Covering stuff.
        self.explicit = kwargs.pop("explicit")
        if self.explicit:
            self.cr_tape_full = kwargs.pop("cr_tape_full")
            self.cr_tape_inner = kwargs.pop("cr_tape_inner")
            self.quad_params = kwargs.pop("quad_params")

            self.inner_inputs = kwargs.pop("inner_inputs")
            self.inner_outputs = kwargs.pop("inner_outputs")
            self.cr_outputs = kwargs.pop("cr_outputs")

            self.tlm_mat_skip_covering = kwargs.pop("tlm_mat_skip_covering", None)
            if self.tlm_mat_skip_covering is None:
                self.tlm_mat_skip_covering = True
        else:
            self.tlm_mat_skip_covering = False

        self.name = kwargs.pop("name", None)
        if self.name is None:
            self.name = "AssembleWithCRBlock"

        # This kwarg will only mess things up.
        kwargs.pop("tensor", None)
        self._assemble_kwargs = kwargs

    def _create_expressions_with_tape_values(self):
        """Replaces original coefficient values with checkpointed values,"""
        replaced_coeffs = {}
        for block_variable in self.get_dependencies():
            coeff = block_variable.output
            c_rep = block_variable.saved_output
            if isinstance(coeff, Expr):
                replaced_coeffs[coeff] = c_rep

        for coeff, out in zip(self._out_terms, self.cr_outputs):
            replaced_coeffs[coeff] = out.block_variable.saved_output

        form = replace(self._form, replaced_coeffs)
        cr_outputs = [r_orig.block_variable.saved_output for r_orig in self.cr_outputs]
        arg_terms = [replace(q, replaced_coeffs) for q in self._arg_terms]
        return form, cr_outputs, arg_terms

    def prepare_evaluate_tlm_matrix(self, inputs, tlm_inputs, relevant_outputs):
        if not self.tlm_mat_skip_covering:
            return super().prepare_evaluate_tlm_matrix(
                inputs, tlm_inputs, relevant_outputs
            )

    def _evaluate_cr_tlm_matrix(self, inputs, relevant_cr_inputs, relevant_cr_outputs):
        """Evaluates the Jacobian of the relevant_outputs for just the CR (no Covering stuff)."""

        # Find the relevant nodes on the tape.
        relevant_cr_input_values = [self.inner_inputs[i] for i in relevant_cr_inputs]
        relevant_cr_input_block_variables = [
            o.block_variable for o in relevant_cr_input_values
        ]
        relevant_cr_output_block_variables = [
            self.cr_outputs[i].block_variable for i in relevant_cr_outputs
        ]

        nodes, blocks = self.tape.find_relevant_nodes(
            relevant_cr_input_block_variables, relevant_cr_output_block_variables
        )
        nodes |= set(relevant_cr_input_block_variables) | set(
            relevant_cr_output_block_variables
        )
        nodes, blocks = self.cr_tape_full.find_relevant_nodes(nodes, nodes)

        # Run inner CR with identity inputs.
        identities = make_jacobian_identities(len(relevant_cr_inputs))
        for bv, identity in zip(relevant_cr_input_block_variables, identities):
            bv.tlm_matrix = identity
        self.cr_tape_inner.evaluate_tlm_matrix(
            inputs=nodes, outputs=nodes, markings=True
        )

        # Extract Jacobian of inner CR, and propagate it to the outer outputs.
        jacobian = []
        for cov_in, cov_out in zip(self.inner_outputs, self.cr_outputs):
            cov_out.block_variable.add_tlm_matrix(cov_in.block_variable.tlm_matrix)
            jacobian.append(cov_in.block_variable.tlm_matrix)

        return jacobian

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        """Evaluates the Jacobian for Pyadjoint.

        Let c be the input variables. We want to calculate dF/dc, which is the
        Jacobian of the form F. Then we can assemble that to get the
        actual desired Jacobian.

        Let r = r(q) be the outputs of the CR and q = q(c) be its inputs. The
        Jacobian dr/dq can be computed separately on its own tape (self.cr_tape_inner)
        in self._evaluate_cr_tlm_matrix.

        The full Jacobian dF/dc = partialF/partialc +  sum_q sum_r (partialF/partialr * dr/dq) * partialq/partialc.
        The partial derivatives can be taken with UFL.

        The return value is a list of Jacobians in the form dF/dc = [dF/dc_1, dF/dc_2, ..., dF/dc_m], where m is
        the number of inputs.
        """
        if not self.tlm_mat_skip_covering:
            return super().evaluate_tlm_matrix_component(
                inputs, tlm_inputs, block_variable, idx, prepared
            )

        F_form, cr_outputs, arg_terms = self._create_expressions_with_tape_values()

        # See which CR outputs are required for the Jacobian.
        relevant_cr_outputs = []
        for i, r in enumerate(cr_outputs):
            if r in F_form.coefficients():
                relevant_cr_outputs.append(i)

        # See which CR inputs are required for the Jacobian.
        relevant_cr_inputs = []
        for i, q in enumerate(arg_terms):
            q_coeffs = ufl.algorithms.analysis.extract_coefficients(q)
            for c_rep, c_jac in zip(inputs, tlm_inputs):
                if (
                    c_jac is not None
                    and isinstance(c_rep, ufl.Coefficient)
                    and c_rep in q_coeffs
                ):
                    relevant_cr_inputs.append(i)
                    break
        drdq_arrays_relevant = self._evaluate_cr_tlm_matrix(
            inputs, relevant_cr_inputs, relevant_cr_outputs
        )

        # Convert the drdq arrays into tensor functions.
        drdq_funcs_relevant = []
        assert len(relevant_cr_outputs) == len(
            drdq_arrays_relevant
        ), f"{len(relevant_cr_outputs)} != {len(drdq_arrays_relevant)}"
        for r_idx, drdq_array_r in zip(relevant_cr_outputs, drdq_arrays_relevant):
            r = cr_outputs[r_idx]
            assert len(relevant_cr_inputs) == len(
                drdq_array_r
            ), f"{len(relevant_cr_inputs)} != {len(drdq_array_r)}"
            drdq_func = []
            for q_idx, drdq_array in zip(relevant_cr_inputs, drdq_array_r):
                if drdq_array is None:
                    drdq_func.append(None)
                    continue
                q = arg_terms[q_idx]
                r_rank, q_rank = len(r.ufl_shape), len(q.ufl_shape)

                # Stick the Jacobian drdq_array into a TensorFunction of the right shape.
                if hasattr(drdq_array, "shape"):
                    if len(drdq_array.shape) > r_rank + q_rank + 1:
                        # Handle the case where the array is the full Jacobian instead of the pointwise-computed Jacobian.
                        drdq_array = np.sum(drdq_array, axis=r_rank + 1)
                    tlm_shape = drdq_array.shape[1:]
                    data = drdq_array.flatten()
                else:
                    # Handle the case where the Jacobian was stored as a Function.
                    tlm_shape = drdq_array.tlm_shape
                    data = drdq_array.vector()[:]

                # Remove 1's from tlm_shape or else the derivative will get mad for shapes not matching.
                tlm_shape = tuple(filter(lambda s: s != 1, tlm_shape))

                T = make_quadrature_space(
                    tlm_shape, self.quad_params, domain=self._form.ufl_domain()
                )

                drdq = backend.Function(T)
                drdq.vector()[:] = data

                drdq_func.append(drdq)
            drdq_funcs_relevant.append(drdq_func)

        # Contract drdq_funcs_relevant with appropriate UFL Jacobians to get desired Jacobian.
        dFdc_all = []
        assert len(inputs) == len(tlm_inputs)
        for c_rep, c_jac in zip(inputs, tlm_inputs):
            if c_jac is None:
                dFdc_all.append(None)
                continue

            c_hat = backend.TrialFunction(c_rep.function_space())

            # First get partial derivative of this coefficient.
            dFdc = derivative(F_form, c_rep, c_hat)

            # Then get derivative contributions from the CRs.
            for r_idx, drdq_r in zip(relevant_cr_outputs, drdq_funcs_relevant):
                r = cr_outputs[r_idx]
                r_rank = len(r.ufl_shape)
                for q_idx, drdq in zip(relevant_cr_inputs, drdq_r):
                    if drdq is None:
                        continue
                    q = arg_terms[q_idx]
                    q_rank = len(q.ufl_shape)

                    # Contract just over the q axes.
                    q_axes = list(range(q_rank))
                    drdq_axes = [a + r_rank for a in q_axes]

                    # Note: this derivative has to be expanded in order for the assembly to not throw an error.
                    dqdc = expand_derivatives(derivative(q, c_rep, c_hat))
                    if q_rank == 0:
                        drdc = drdq * dqdc
                    else:
                        drdc = contraction(drdq, drdq_axes, dqdc, q_axes)
                    dFdc += derivative(F_form, r, drdc)

            J = backend.assemble(
                dFdc, form_compiler_parameters=self.quad_params, **self._assemble_kwargs
            )
            dFdc_all.append(J)

        # Contract dFdc_all with the tlm_inputs to get dFdx_all.
        dcdx_all = tlm_inputs
        dFdx_all = None
        for i, (dFdc, dcdx) in enumerate(zip(dFdc_all, dcdx_all)):
            if dcdx is None:
                continue
            if dFdx_all is None:
                dFdx_all = [None] * len(dcdx)

            for j, dcdx_j in enumerate(dcdx):
                # Add dFdc @ dc_dx_j to dFdx_all[j].
                if isinstance(dcdx_j, JacobianIdentity):
                    dot = dFdc
                else:
                    raise ValueError("Cannot handle non-identity inputs")
                if dFdx_all[j] is None:
                    dFdx_all[j] = dot
                else:
                    dFdx_all[j] += dot

        return dFdx_all



from pyadjoint.enlisting import Enlist
from ufl import Form
from ufl.algorithms.estimate_degrees import SumDegreeEstimator as origSumDegreeEstimator
from ufl.corealg.map_dag import map_expr_dags
from ufl.integral import Integral
from ..fe import mesh_from_ufl_domain, FunctionSpace, FiniteElement, TensorElement


def get_quadrature_params(quad_params=None, default_degree=1):
    if quad_params is None:
        quad_params = {}
    degree = quad_params.get("quadrature_degree", default_degree)
    scheme = quad_params.get("quadrature_rule", "default")

    quad_params["quadrature_rule"] = scheme
    quad_params["quadrature_degree"] = degree
    quad_params["representation"] = "quadrature"
    return quad_params


def make_quadrature_spaces(spaces, quad_params=None, domain=None):
    # Note: Using these points to estimate the quadrature degree may be almost
    #       useless. They're probably not representative of the degree of the
    #       expressions that will be computed.
    spaces = Enlist(spaces)
    points = [s.point() for s in spaces]
    max_arg_degree = max(tuple(estimate_total_polynomial_degree(p) for p in points))
    quad_params = get_quadrature_params(quad_params, max_arg_degree)

    n = len(spaces)
    quad_spaces = [None] * n
    for i in range(n):
        quad_spaces[i] = make_quadrature_space(
            spaces[i].shape(), quad_params, expr=points[i], domain=domain
        )
    return spaces.delist(tuple(quad_spaces)), quad_params


def make_quadrature_space(shape, quad_params, expr=None, domain=None):
    degree = quad_params["quadrature_degree"]
    scheme = quad_params.get("quadrature_rule", "default")
    domain = domain if domain is not None else expr.ufl_domain()
    if domain is None:
        raise ValueError(
            "You must specify the domain (or an expression on the domain) to create a quadrature space"
        )
    cell = domain.ufl_cell()
    mesh = mesh_from_ufl_domain(domain)

    if shape:
        TE = TensorElement(
            "Quadrature", cell, degree=degree, shape=shape, quad_scheme=scheme
        )
    else:
        TE = FiniteElement("Quadrature", cell, degree=degree, quad_scheme=scheme)
    TV = FunctionSpace(mesh, TE)
    return TV


class SumDegreeEstimator(origSumDegreeEstimator):
    """
    This class is equivalent to the original SumDegreesEstimator class except
    for the addition of the of coefficient_replace_map. This map allows the
    degree of any coefficient to be explicitly specified.

    This class also adds handling for sym. I'm not sure if it's correct, though.
    """

    def __init__(self, default_degree, element_replace_map, coefficient_replace_map):
        origSumDegreeEstimator.__init__(self, default_degree, element_replace_map)
        self.coefficient_replace_map = coefficient_replace_map

    def coefficient(self, v):
        """A form coefficient provides a degree based on the coefficient map,
        or the element map, or the default degree if the element has no degree."""
        v = self.coefficient_replace_map.get(v, v)
        if isinstance(v, int):
            return v
        return super().coefficient(v)

    sym = origSumDegreeEstimator._add_degrees


def estimate_total_polynomial_degree(
    e, default_degree=1, element_replace_map={}, coefficient_replace_map={}
):
    """This is equivalent to the original estimate_total_polynomial_degree
    except it supports the coefficient_replace_map argument so that the degree
    of any coefficient can be explicitly specified."""
    de = SumDegreeEstimator(
        default_degree, element_replace_map, coefficient_replace_map
    )
    if isinstance(e, Form):
        if not e.integrals():
            error("Got form with no integrals!")
        degrees = map_expr_dags(de, [it.integrand() for it in e.integrals()])
    elif isinstance(e, Integral):
        degrees = map_expr_dags(de, [e.integrand()])
    else:
        degrees = map_expr_dags(de, [e])
    degree = max(degrees) if degrees else default_degree
    return degree



from pyadjoint.overloaded_type import register_overloaded_type
from pyadjoint.adjfloat import AdjFloat
import numpy
from .array import ndarray
from .autograd import overload_autograd

register_overloaded_type(AdjFloat, int)
register_overloaded_type(AdjFloat, numpy.float64)
register_overloaded_type(AdjFloat, numpy.float32)
register_overloaded_type(AdjFloat, numpy.int64)
register_overloaded_type(AdjFloat, numpy.int32)



from pyadjoint.block_variable import BlockVariable as pyadjoint_block_variable


class BlockVariable(pyadjoint_block_variable):
    """References a block output variable."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tlm_matrix = None

    def add_tlm_matrix(self, val):
        if self.tlm_matrix is None:
            self.tlm_matrix = val
        else:
            self.tlm_matrix += val

    def reset_variables(self, types):
        super().reset_variables(types)

        if "tlm_matrix" in types:
            self.tlm_matrix = None



from pyadjoint import set_working_tape
from pyadjoint_utils import Tape

set_working_tape(Tape())
del Tape
del set_working_tape

from .utils import homogenize_bcs
from .assembly import assemble, AssembleBlock
from .array import function_get_local, function_set_local

from fenics_adjoint import backend, compat

if backend.__name__ == "fenics":
    # Fix for https://bitbucket.org/dolfin-adjoint/pyadjoint/issues/118/fenics-compatassemble_adjoint_value-throws
    def assemble_adjoint_value(*args, **kwargs):
        """Wrapper that assembles a matrix with boundary conditions"""
        bcs = kwargs.pop("bcs", ())
        result = backend.assemble(*args, **kwargs)
        for bc in bcs:
            bc.apply(result)
        return result

    compat.assemble_adjoint_value = assemble_adjoint_value
    del assemble_adjoint_value



import numpy
from numpy_adjoint.array import ndarray as orig_ndarray
from pyadjoint.overloaded_type import (
    OverloadedType,
    register_overloaded_type,
    create_overloaded_object,
)
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint_utils.block import Block

# The official numpy_adjoint branch is still in development (as of version 2019.1.0), so in the mean
# time, I'm overwriting/adding the basic methods I need to be able to use numpy arrays as Pyadjoint
# Controls.


class ndarray(orig_ndarray):
    @classmethod
    def _ad_init_object(cls, obj):
        obj = numpy.array(obj, subok=True)
        return cls(obj.shape, dtype=obj.dtype, buffer=obj.data, strides=obj.strides)

    def _ad_dim(self):
        return self.size

    def _ad_dot(self, other):
        return numpy.dot(self.flat, other.flat)

    def _ad_mul(self, other):
        return self * other

    def _ad_add(self, other):
        return self + other

    def _ad_copy(self):
        return self.copy()

    def copy(self, *args, **kwargs):
        return create_overloaded_object(super().copy(*args, **kwargs))

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        dst = numpy.reshape(src[offset : offset + dst.size], dst.shape)
        offset += dst.size
        return dst, offset

    @staticmethod
    def _ad_to_list(self):
        return self.flatten().tolist()

    def __getitem__(self, item):
        annotate = annotate_tape()
        if annotate:
            block = NumpyArraySliceBlock(self, item)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            out = numpy.ndarray.__getitem__(self, item)

        if annotate:
            out = create_overloaded_object(out)
            block.add_output(out.create_block_variable())
        return out


class NumpyArraySliceBlock(Block):
    def __init__(self, array, item):
        super().__init__()
        self.add_dependency(array)
        self.item = item

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        adj_output = numpy.zeros(inputs[0].shape)
        adj_output[self.item] = adj_inputs[0]
        return adj_output

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return inputs[0][self.item]

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return tlm_inputs[0][self.item]


register_overloaded_type(ndarray, numpy.ndarray)



from fenics import *
from fenics_adjoint import *
from pyadjoint import *
from pyadjoint_utils import *
from ..network_block import run_network
from .space_builders import DirectSum
from .ufl import UFLFunctionSpace
from .types import PointMap
import numpy as np


def get_total_value_dimension(spaces):
    total = 0
    for V in spaces:
        total += V.element().value_dimension(0)
    return total


class CRNetwork(PointMap):
    def __init__(self, input_spaces, output_space, network):
        self.input_spaces = input_spaces
        self.output_space = output_space

        self.num_inputs = get_total_value_dimension(self.input_spaces)
        self.num_outputs = get_total_value_dimension((self.output_space,))

        # Make sure the network has the right number of inputs and outputs.
        self.setParams([network])
        assert (
            self.network.num_inputs == self.num_inputs
        ), "Network has %d inputs but should have %d" % (
            self.network.num_inputs,
            self.num_inputs,
        )
        assert (
            self.network.num_outputs == self.num_outputs
        ), "Network has %d outputs but should have %d" % (
            self.network.num_outputs,
            self.num_outputs,
        )

        arg_space = DirectSum([UFLFunctionSpace(V) for V in self.input_spaces])
        out_space = UFLFunctionSpace(self.output_space)

        if len(arg_space) == 1:
            arg_space = arg_space[0]

        super().__init__(arg_space, out_space)

    def __call__(self, args):
        if not isinstance(self.source, DirectSum):
            args = [args]
        output = run_network(self.network, self.output_space, *args)

        self.inputs = args
        self.controls = [Control(a) for a in self.inputs]
        self.rf = ReducedFunction(output, self.controls)
        return output

    def numpy_call(self, *inputs):
        net_input = np.concatenate(inputs, axis=1)
        return self.network.predict(net_input)

    def setParams(self, params):
        if not isinstance(params, (list, tuple)):
            params = [params]
        self.network = params[0]
        self.params = [self.network]
        self.params_controls = [Control(self.network)]



from fenics import plot as backend_plot
import os
import re

try:
    import matplotlib.pyplot as plt
except:
    pass

# Just use `debug(locals(), globals())` to drop into an interactive Python interpreter.
def debug(locals, globals, msg="(InteractiveConsole)"):
    import code

    code.interact(
        local=dict(globals, **locals),
        banner="\n" + msg,
        exitmsg="(Exiting InteractiveConsole)\n",
    )


## add_bool_arg() adds a boolean argument to the given argparse parser.
def add_bool_arg(parser, name, default=False, help=""):
    group = parser.add_mutually_exclusive_group(required=False)
    if default:
        group.add_argument("--no-" + name, dest=name, action="store_false", help=help)
    else:
        group.add_argument("--" + name, dest=name, action="store_true", help=help)
    parser.set_defaults(**{name: default})


## plot() plots the given function, adds a legend for the color gradient, and adds a new figure
#  so that subsequent plot() calls don't overwrite this plot. Pass newfig=False if you don't
#  want it to create a new figure afterwards.
FIGURE_SAVE_DIR = False
SHOW_FIGURES = True


def plot(u, **kwargs):
    newfig = kwargs.pop("newfig", True)
    show = kwargs.pop("noshow", SHOW_FIGURES)
    try:
        if hasattr(u, "shape") and (
            len(u.shape) == 2 or (len(u.shape) == 3 and u.shape[2] in (3, 4))
        ):
            # Plot matrix.
            f = plt.imshow(u, **kwargs)
        else:
            f = backend_plot(u, **kwargs)
        try:
            plt.colorbar(f)
        except:
            pass
        saveplot()
        if not show:
            plt.close()
        if newfig:
            plt.figure()
        return f
    except Exception as e:
        print("Warning: couldn't plot stuff:", e)


def saveplot(filename=None, **kwargs):
    savedir = kwargs.pop("savedir", FIGURE_SAVE_DIR)
    if savedir:
        fig = plt.gcf()
        os.makedirs(savedir, exist_ok=True)

        # Remove dots from filename so that matplotlib doesn't interpret it as image format.
        if filename is None:
            filename = fig.gca().get_title()
        filename = re.sub(r"\.", "_", filename)
        fig.savefig(savedir + "/" + filename)


## showplot() closes the current plot (unless you pass closefig=False) and then calls plt.show()
def showplot(closefig=True):
    try:
        if closefig:
            plt.close()
        plt.show()
    except Exception as e:
        print("Warning: couldn't show plot:", e)


def print_dir(o, private=False):
    d = dir(o)
    if not private:
        d = filter(lambda s: not s.startswith("_"), d)
    print("\n".join(d))



from .utils import (
    symm,
    antisymm,
    commutator_action,
    anticommutator_action,
    scalar_triple_prod,
    powerset,
    eps_ij,
    eps_ijk,
    axial_vector,
    near,
    levi_civita,
)
from .invariants import (
    get_invariant_functions,
    get_invariant_descriptions,
    InvariantInfo,
    TensorType,
    LeviCivitaType,
    register_invariant_functions,
    type_from_array,
)



import logging
from fenics import set_log_level as _backend_set_log_level

logger = logging.getLogger("CRIKit")
"""
The logger for CRIKit. To set the log level, after importing ``crikit`` or any object contained in it, simply use the ``logging`` module like

.. testcode::

     import crikit
     crikit.logging.set_log_level(crikit.logging.WARNING)

"""

DEBUG: int = 10
TRACE: int = 13
PROGRESS: int = 16
INFO: int = 20
WARNING: int = 30
ERROR: int = 40
CRITICAL: int = 50


def set_log_level(level):
    """
        Set the log level for CRIKit and FEniCS. For example,

    .. testcode::

         import crikit
         crikit.logging.set_log_level(crikit.logging.WARNING)

    """
    logger.setLevel(level)
    _backend_set_log_level(level)



from pyadjoint import Control as pyadjoint_control


class Control(pyadjoint_control):
    @property
    def tlm_matrix(self):
        return self.block_variable.tlm_matrix

    @tlm_matrix.setter
    def tlm_matrix(self, value):
        self.block_variable.tlm_matrix = value



from pyadjoint import Tape as pyadjoint_tape
from pyadjoint import get_working_tape, set_working_tape, Block

from contextlib import contextmanager


@contextmanager
def push_tape():
    """Creates a new tape in its scope that is a sub-tape of the current working tape"""
    orig_tape = get_working_tape()

    new_tape = Tape()
    set_working_tape(new_tape)
    yield new_tape

    set_working_tape(orig_tape)


def _find_relevant_nodes(tape, controls_or_block_variables):
    # This function is just a stripped down Block.optimize_for_controls
    blocks = tape.get_blocks()
    if len(controls_or_block_variables) > 0 and hasattr(
        next(iter(controls_or_block_variables)), "block_variable"
    ):
        nodes = set([control.block_variable for control in controls_or_block_variables])
    else:
        nodes = set([block_variable for block_variable in controls_or_block_variables])

    for block in blocks:
        depends_on_control = False
        for dep in block.get_dependencies():
            if dep in nodes:
                depends_on_control = True

        if depends_on_control:
            for output in block.get_outputs():
                nodes.add(output)
    return nodes


def _block_get_tf_blocks(block):
    if not hasattr(block, "tf_get_blocks"):
        return block
    tf_blocks = block.tf_get_blocks()
    if isinstance(tf_blocks, Block):
        return tf_blocks

    block, sub_blocks = tf_blocks
    sub_blocks = list(sub_blocks)
    for i, sub_block in enumerate(sub_blocks):
        sub_blocks[i] = _block_get_tf_blocks(sub_block)
    return [block, sub_blocks]


class Tape(pyadjoint_tape):

    __slots__ = ["_tf_output_block_lookup"]

    def __init__(self, *args, **kwargs):
        self._tf_output_block_lookup = {}
        super().__init__(*args, **kwargs)

    def evaluate_adj(self, inputs=None, outputs=None, markings=False):
        nodes, blocks = self.find_relevant_nodes(inputs, outputs)
        for block in reversed(blocks):
            block.evaluate_adj(markings=markings)

    def evaluate_tlm(self, inputs=None, outputs=None, markings=False):
        nodes, blocks = self.find_relevant_nodes(inputs, outputs)
        with self.marked_nodes(nodes):
            for block in blocks:
                block.evaluate_tlm(markings=True)

    def evaluate_tlm_matrix(self, inputs=None, outputs=None, markings=False):
        nodes, blocks = self.find_relevant_nodes(inputs, outputs)
        with self.marked_nodes(nodes, find_outputs=False):
            for block in blocks:
                block.evaluate_tlm_matrix(markings=True)

    def evaluate_hessian(self, inputs=None, outputs=None, markings=False):
        nodes, blocks = self.find_relevant_nodes(inputs, outputs)
        for block in reversed(blocks):
            block.evaluate_hessian(markings=markings)

    def recompute(self, inputs=None, outputs=None):
        nodes, blocks = self.find_relevant_nodes(inputs, outputs)
        for block in blocks:
            block.recompute()
            # print(f"output of block {block} is {block.get_outputs()}")

    def reset_tlm_matrix_values(self):
        for block in reversed(self._blocks):
            block.reset_variables(types=("tlm_matrix"))

    def find_relevant_dependencies(self, outputs):
        blocks = self.get_blocks()
        if len(outputs) > 0 and hasattr(next(iter(outputs)), "block_variable"):
            nodes = set([output.block_variable for output in outputs])
        else:
            nodes = set(outputs)
        relevant_blocks = [False] * len(blocks)

        for i in range(len(blocks) - 1, -1, -1):
            block = blocks[i]
            produces_output = False
            for dep in block.get_outputs():
                if dep in nodes:
                    produces_output = True

            relevant_blocks[i] = produces_output
            if produces_output:
                for dep in block.get_dependencies():
                    nodes.add(dep)
        return nodes, relevant_blocks

    def find_relevant_outputs(self, inputs):
        blocks = self.get_blocks()
        if len(inputs) > 0 and hasattr(next(iter(inputs)), "block_variable"):
            nodes = set([control.block_variable for control in inputs])
        else:
            nodes = set(inputs)
        relevant_blocks_mask = []

        for block in blocks:
            depends_on_control = False
            for dep in block.get_dependencies():
                if dep in nodes:
                    depends_on_control = True

            relevant_blocks_mask.append(depends_on_control)
            if depends_on_control:
                for output in block.get_outputs():
                    nodes.add(output)
        return nodes, relevant_blocks_mask

    def find_relevant_nodes(self, inputs=None, outputs=None):
        # TODO: double check that inputs and outputs are block variables or controls.
        blocks = self.get_blocks()

        if inputs is None:
            if outputs is None:
                nodes = set()
                for block in blocks:
                    for dep in block.get_dependencies():
                        nodes.add(dep)
                    for output in block.get_outputs():
                        nodes.add(output)
                return nodes, blocks
            else:
                nodes, relevant_blocks_masks = self.find_relevant_dependencies(outputs)
                relevant_blocks = [
                    block for i, block in enumerate(blocks) if relevant_blocks_masks[i]
                ]
                return nodes, relevant_blocks
        else:
            if outputs is None:
                nodes, relevant_blocks_masks = self.find_relevant_outputs(inputs)
                relevant_blocks = [
                    block for i, block in enumerate(blocks) if relevant_blocks_masks[i]
                ]
                return nodes, relevant_blocks
            else:
                i_nodes, i_blocks = self.find_relevant_outputs(inputs)
                o_nodes, o_blocks = self.find_relevant_dependencies(outputs)

                nodes = i_nodes.intersection(o_nodes)
                relevant_blocks_masks = [i and o for i, o in zip(i_blocks, o_blocks)]
                relevant_blocks = [
                    block for i, block in enumerate(blocks) if relevant_blocks_masks[i]
                ]
                return nodes, relevant_blocks

    def find_absolute_dependencies_outputs(self):
        dependencies = set()
        outputs = set()
        for block in self._blocks:
            for dep in block.get_dependencies():
                if dep not in outputs:
                    dependencies.add(dep)
            for output in block.get_outputs():
                outputs.add(output)
        return dependencies, outputs

    # This function was modified to work directly with block variables.
    @contextmanager
    def marked_nodes(self, controls_or_block_variables, find_outputs=True):
        if find_outputs:
            nodes = _find_relevant_nodes(self, controls_or_block_variables)
        else:
            nodes = controls_or_block_variables
        old_values = [node.marked_in_path for node in nodes]
        for node in nodes:
            node.marked_in_path = True
        yield
        for node, old_value in zip(nodes, old_values):
            node.marked_in_path = old_value

    @contextmanager
    def save_adj_values(self):
        nodes, _ = self.find_relevant_nodes()
        old_values = [node.adj_value for node in nodes]
        for node in nodes:
            node.adj_value = None
        yield
        for node, old_value in zip(nodes, old_values):
            node.adj_value = old_value

    def _get_tf_scope_name(self, node):
        """Return a TensorFlow scope name based on the node's class name or an attribute 'tf_name'."""
        # If the block is a BlockVariable we use block.output
        if node.__class__.__name__ == "BlockVariable":
            node = node.output
        if hasattr(node, "tf_name"):
            name = node.tf_name
        else:
            name = node.__class__.__name__
        return self._valid_tf_scope_name(name)

    def _tf_register_blocks(self, name=None):
        lst = [name]
        for block in self.get_blocks():
            if block in self._tf_added_blocks:
                continue
            self._tf_added_blocks.append(block)
            lst.append(_block_get_tf_blocks(block))
        self._tf_registered_blocks.append(lst)

    def _tf_rebuild_registered_blocks(self):
        """Remove blocks that no longer exist on the tape from registered blocks."""
        new_registered_blocks = []
        new_added_blocks = []
        for scope in self._tf_registered_blocks:
            lst = scope[:1]
            for block in scope[1:]:
                if isinstance(block, Block):
                    if block in self.get_blocks():
                        lst.append(block)
                else:
                    block, sub_blocks = block
                    if block in self.get_blocks():
                        lst.append(_block_get_tf_blocks(block))
                new_added_blocks.append(block)
            if len(lst) > 1:
                new_registered_blocks.append(lst)
        self._tf_registered_blocks = new_registered_blocks
        self._tf_added_blocks = new_added_blocks

    def _tf_add_blocks_scoped(self, blocks):
        """Add the given blocks (with possible sub-blocks) to the TensorFlow graph."""

        import tensorflow.compat.v1 as tf

        for block in blocks:
            self._tf_add_block(block)

        for block in blocks:
            if not isinstance(block, Block):
                block = block[0]
            for out in block.get_outputs():
                if id(out) not in self._tf_tensors:
                    t = self._tf_output_block_lookup.get(id(out), None)
                    if t is None:
                        # This block output wasn't created in the tensorflow
                        # graph, and I have no way to create it because I don't
                        # know what block generated the corresponding tensor.
                        raise ValueError(
                            "This output is expected to be created in a sub-block: {}".format(
                                str(out)
                            )
                        )
                    with tf.name_scope(self._get_tf_scope_name(out)):
                        tout = tf.py_function(
                            lambda: None,
                            [t],
                            [tf.float64],
                            name=self._valid_tf_scope_name(str(out)),
                        )
                    self._tf_tensors[id(out)] = tout

    def visualise(self, output="log", *args, **kwargs):
        """This resets the TensorFlow data, which allows a user to call tape.visualise() twice without error."""
        if not output.endswith(".dot"):
            self._tf_tensors = {}
            self._tf_added_blocks = []
            self._tf_registered_blocks = []
        return super().visualise(output, *args, **kwargs)

    def _tf_add_blocks(self):
        """Add new blocks to the TensorFlow graph while supporting the name_scope() method."""

        import tensorflow as tf

        self._tf_register_blocks()
        self._tf_rebuild_registered_blocks()

        for scope in self._tf_registered_blocks:
            scope_name = scope[0]
            if scope_name is None:
                self._tf_add_blocks_scoped(scope[1:])
            else:
                with tf.name_scope(scope_name):
                    self._tf_add_blocks_scoped(scope[1:])

    def _tf_add_block(self, block, sub_blocks=None):
        """Add a block to the TensorFlow graph, and recursively add its sub-blocks."""

        import tensorflow as tf

        if not isinstance(block, Block):
            if sub_blocks is not None:
                raise ValueError(
                    "If the input is not a Block, then it should be a tuple (block, sub_blocks) and the sub_blocks kwarg must be None"
                )
            block, sub_blocks = block

        # Block dependencies
        in_tensors = []
        for dep in block.get_dependencies():
            if id(dep) in self._tf_tensors:
                in_tensors.append(self._tf_tensors[id(dep)])
            else:
                # Look up the block variable in _tf_output_block_lookup to
                # connect it to a previously recorded block if necessary.
                inputs = self._tf_output_block_lookup.get(id(dep), [])
                with tf.name_scope(self._get_tf_scope_name(dep)):
                    tin = tf.numpy_function(
                        lambda: None,
                        inputs,
                        [tf.float64],
                        name=self._valid_tf_scope_name(str(dep)),
                    )
                    in_tensors.append(tin)
                    self._tf_tensors[id(dep)] = tin

        # Block node
        with tf.name_scope(self._get_tf_scope_name(block)):
            if sub_blocks is None:

                def tf_np_f(*args):
                    return None

                tensor = tf.numpy_function(
                    tf_np_f,
                    in_tensors,
                    [tf.float64],
                    name=self._valid_tf_scope_name(str(block)),
                )
                self._tf_tensors[id(block)] = tensor
            else:
                for sub_block in sub_blocks:
                    self._tf_add_block(sub_block)
            if hasattr(block, "tf_add_extra_to_graph"):
                block.tf_add_extra_to_graph(self._tf_tensors)

        # Block outputs.
        # To avoid incorrectly scoping the block outputs, these will be added
        # once they are used as input. The output tensor of the current block must
        # be saved to be used as input when the block outputs are created.
        if sub_blocks is None:
            for out in block.get_outputs():
                self._tf_output_block_lookup[id(out)] = tensor



import crikit.network as backend_network
import numpy as np
from pyadjoint.overloaded_type import (
    OverloadedType,
    create_overloaded_object,
    register_overloaded_type,
)


@register_overloaded_type
class Network(OverloadedType, backend_network.Network):
    def __init__(self, *args, **kwargs):
        super(Network, self).__init__(*args, **kwargs)
        backend_network.Network.__init__(self, *args, **kwargs)

    def create_copy(self, weights=None):
        return create_overloaded_object(
            backend_network.Network.create_copy(self, weights)
        )

    @classmethod
    def _ad_init_object(cls, obj):
        r = cls(obj.layers)
        return r

    def _create_from_values(self, values):
        return self.create_copy(weights=values)

    # This takes the result of an adjoint computation and converts it to my type.
    def _ad_convert_type(self, value, options={}):
        if value is None:
            # TODO: Should the default be 0 constant here or return just None?
            return None
        if isinstance(value, backend_network.Network):
            return value
        return self._create_from_values(value)

    def _ad_create_checkpoint(self):
        return self.create_copy()

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def _ad_mul(self, other):
        weights = self.get_weights()
        for w in weights:
            w *= other
        return self._create_from_values(weights)

    def __add__(self, other):
        return self._ad_add(other)

    def _ad_add(self, other):
        weights = self.get_weights()
        other_weights = other.get_weights(copy=False)
        for w, ow in zip(weights, other_weights):
            w += ow
        return self._create_from_values(weights)

    def _ad_dot(self, other, options=None):
        dot = 0
        weights = self.get_weights(copy=False)
        other_weights = other.get_weights(copy=False)
        for w, ow in zip(weights, other_weights):
            dot += np.sum(w * ow)
        return dot

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        weights = dst.get_weights(copy=False)
        new_weights = []
        for w in weights:
            new_weights.append(np.reshape(src[offset : offset + w.size], w.shape))
            offset += w.size
        dst.set_weights(new_weights)
        return dst, offset

    @staticmethod
    def _ad_to_list(m):
        weights = m.get_weights(copy=False)
        flattened_weights = []
        for w in weights:
            for row in w:
                for val in row:
                    flattened_weights.append(val)
        return flattened_weights

    def _ad_copy(self):
        return self.create_copy()

    def _ad_dim(self):
        weights = self.get_weights(copy=False)
        dim = 0
        for w in weights:
            dim += w.size
        return dim


@register_overloaded_type
class PlapNetwork(OverloadedType, backend_network.PlapNetwork):
    def __init__(self, *args, **kwargs):
        super(PlapNetwork, self).__init__(*args, **kwargs)
        backend_network.PlapNetwork.__init__(self, *args, **kwargs)

    def create_copy(self, weights=None):
        # TODO: this is inefficient because it creates a backend copy and then
        # create_overloaded_object calls _ad_init_object, which creates another copy.
        return create_overloaded_object(
            backend_network.PlapNetwork.create_copy(self, weights)
        )

    @classmethod
    def _ad_init_object(cls, obj, weights=None):
        p = obj.p if weights is None else weights
        r = cls(
            p,
            dims=obj.dims,
            input_just_vector=obj.input_just_vector,
            output_vector=obj.output_vector,
        )
        return r

    # This takes the result of an adjoint computation and converts it to my type.
    def _ad_convert_type(self, value, options={}):
        if value is None:
            # TODO: Should the default be 0 constant here or return just None?
            return None
        if isinstance(value, backend_network.PlapNetwork):
            return value
        return self.create_copy(value)

    def _ad_create_checkpoint(self):
        return self.create_copy(self.p)

    def _ad_restore_at_checkpoint(self, checkpoint):
        return checkpoint

    def _ad_mul(self, other):
        return self.create_copy(self.p * other)

    def __add__(self, other):
        return self._ad_add(other)

    def _ad_add(self, other):
        return self.create_copy(self.p + other.p)

    def _ad_dot(self, other, options=None):
        return self.p * other.p

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        dst.p = src[offset]
        offset += 1
        return dst, offset

    @staticmethod
    def _ad_to_list(m):
        return [m.p]

    def _ad_copy(self):
        return self.create_copy(self.p)

    def _ad_dim(self):
        return 1



"""This submodule defines a default FE adjoint backends (FEniCS or Firedrake)"""

from .fe import crikit_fe_backend

if crikit_fe_backend == "firedrake":
    from firedrake_adjoint import *
elif crikit_fe_backend == "fenics":
    from fenics_adjoint import *
else:
    ImportError("Unrecognized adjoint backend")

from pyadjoint_utils import *



import jax
import jax.numpy as jnp
from pyadjoint_utils.jax_adjoint import ndarray, array
from pyadjoint import AdjFloat
from pyadjoint.enlisting import Enlist
from pyadjoint_utils import Block, JacobianIdentity
from pyadjoint import Control
from pyadjoint.overloaded_type import (
    OverloadedType,
    register_overloaded_type,
    create_overloaded_object,
)
from pyadjoint.overloaded_function import overload_function
from pyadjoint.tape import annotate_tape, get_working_tape, stop_annotating
from functools import wraps
from jax.tree_util import Partial as partial  # JAX-friendlier functools.partial
from jax.tree_util import tree_flatten, tree_unflatten, tree_leaves, tree_map
from itertools import product
from typing import Tuple, Optional, Callable, Iterable, Sequence, Union, Any, TypeVar
from .array import convert_arg
import logging

logger = logging.getLogger("CRIKit")

# These TypeVars express the fact that
# the overload_jax() function is a transformation
# from a general function to a pyadjoint-overloaded function
# (i.e. one that can take in OverloadedObject instances as inputs
# and return them as outputs, correctly taping the operation they represent)
Function = TypeVar("Function", bound=Callable)
OverloadedFunction = TypeVar("OverloadedFunction", bound=Callable)


class JAXBlock(Block):
    def __init__(
        self,
        func,
        args,
        outputs,
        argnums=None,
        nojit=False,
        jit_kwargs=None,
        pointwise=False,
        out_pointwise=None,
        compile_jacobian_stack=True,
        **kwargs,
    ):
        super().__init__()
        self._func_name = kwargs.get("func_name", None) or func.__name__
        self._holomorphic = kwargs.get("holomorphic", False)
        self._func = func
        self._differentiable_func = self._diff_func
        self._nojit = nojit
        self._jit_kwargs = jit_kwargs if jit_kwargs is not None else {}
        self._compiled_jacobian_stack = compile_jacobian_stack
        num_inputs = len(args)
        num_outputs = len(outputs)
        if argnums:
            self._tlm = jax.jacrev(
                self._differentiable_func,
                argnums=argnums,
                holomorphic=self._holomorphic,
            )
        else:
            self._tlm = jax.jacrev(
                self._differentiable_func, holomorphic=self._holomorphic
            )
        self._jvp_maker = _pushforward(self._differentiable_func)

        argnum_range = tuple(argnums) if argnums else range(num_inputs)
        self._tlm_argnums = argnum_range
        self._tlm_outnums = tuple(range(num_outputs))
        self._tlm_matrix_func = self._jit(
            jax.jacrev(self._diff_func, argnums=self._tlm_argnums)
        )
        self._vjpfuncs = [
            vector_jacobian_product(self._func, i, jitfun=self._jit)
            for i in argnum_range
        ]
        self._hesfuncs = [
            vector_jacobian_product(self._vjpfuncs[i], j, False, jitfun=self._jit)
            for i, j in product(range(len(argnum_range)), argnum_range)
        ]

        iarange = range(len(argnum_range))
        self._make_hvpfun = _make_hvpfun(self._differentiable_func)
        self._hessian = jax.hessian(
            self._differentiable_func, argnums=range(len(argnum_range))
        )
        self._dependencies = []
        self._outputs = []
        self._param_range = argnum_range

        # flatten args and add all dependencies. The input signature of the function is
        # typically something like ((arg1,arg2,...),param1,param2,...), so JAX's pytree
        # utilities are helpful for this stuff
        arg_leaves = []
        for a in args:
            if isinstance(a, (tuple, list)):
                for v in a:
                    arg_leaves.append(v)
            else:
                arg_leaves.append(a)

        if len(self._param_range) > len(arg_leaves):
            raise ValueError(
                f"There are more argument indices in argnums than there are arguments ({len(self._param_range)} > {len(arg_leaves)})"
            )
        if max(self._param_range) >= len(arg_leaves):
            raise ValueError(
                f"There is an argument index in argnums that is higher than the number of arguments ({max(self._param_range)} >= {len(arg_leaves)})"
            )

        # Record dependencies only for the specified arguments.
        for idx in self._param_range:
            self.add_dependency(arg_leaves[idx])

        leaves, self._treedef = tree_flatten(tuple(args))

        self._args = [convert_arg(x) for x in leaves]

        for out in outputs:
            self.add_output(out.create_block_variable())

        self._saved_outs = outputs
        self._single_output = num_outputs == 1

        # Handle pointwise specifications.
        self._in_pointwise = (
            (pointwise,) * num_inputs if isinstance(pointwise, bool) else pointwise
        )
        self._any_pointwise = any(self._in_pointwise)
        if out_pointwise is None:
            out_pointwise = self._any_pointwise
        self._out_pointwise = (
            (out_pointwise,) * num_outputs
            if isinstance(out_pointwise, bool)
            else out_pointwise
        )
        if len(self._in_pointwise) != num_inputs:
            raise ValueError(
                f"len(pointwise) != num_inputs. For each input, you must specify if it is defined pointwise. ({len(self._in_pointwise)} != {num_inputs})"
            )
        if len(self._out_pointwise) != num_outputs:
            raise ValueError(
                f"len(pointwise) != num_outputs. For each output, you must specify if it is defined pointwise. ({len(self._out_pointwise)} != {num_outputs})"
            )
        if any(self._in_pointwise) and not any(self._out_pointwise):
            raise ValueError(
                "There is a pointwise input and no pointwise outputs. If an input is specified as pointwise, then at least one output must be specified as pointwise."
            )
        if any(self._out_pointwise) and not any(self._in_pointwise):
            raise ValueError(
                "There is a pointwise output and no pointwise inputs. If an output is specified as pointwise, then at least one input must be specified as pointwise."
            )

        self._compiled_jacobian = False

    def _jit(self, func):
        return func if self._nojit else jax.jit(func, **self._jit_kwargs)

    def _diff_func(self, *args):
        leaves, treedef = tree_flatten(tuple(args))
        self._replace_params(leaves)
        argtree = tree_unflatten(self._treedef, self._args)
        args = tuple(convert_arg(x) for x in argtree)
        return self._func(*args)

    def _make_partial_diff_func(self, idxs):
        def _pdiff_func(*args):
            leaves, treedef = tree_flatten(tuple(args))
            param_idxs = [self._param_range[idx] for idx in idxs]
            self._replace_params_partial(args, param_idxs)
            argtree = tree_unflatten(self._treedef, self._args)
            args = tuple(convert_arg(x) for x in argtree)
            return self._func(*args)

        return _pdiff_func

    def _replace_params(self, new_params):
        for new_param, idx in zip(new_params, self._param_range):
            par = convert_arg(new_param)
            self._args[idx] = par

    def _replace_params_partial(self, new_params, idxs):
        for i, idx in enumerate(idxs):
            self._args[idx] = convert_arg(new_params[i])

    def _get_reduced_output_diff_func(self, output_ids):
        if self._single_output:
            return self._diff_func

        def diff_func_relevant(*args):
            outputs = self._diff_func(*args)
            outputs = tuple(outputs[idx] for idx in output_ids)
            return outputs

        return diff_func_relevant

    def _add_params(self, deltas):
        for i, idx in enumerate(self._param_range):
            self._args[idx] += convert_arg(deltas[i])

    def __repr__(self):
        return f"JAXBlock({self._func_name})"

    def prepare_recompute_component(self, inputs, relevant_outputs):
        return Enlist(self._diff_func(*inputs))

    def recompute_component(self, inputs, block_variable, idx, prepared):
        if prepared:
            return prepared[idx]
        elif self._saved_outs:
            return self._saved_outs[idx]
        else:
            self._replace_params(inputs)
            self._saved_outs = Enlist(self._differentiable_func(*self._args))
            return self._saved_outs[idx]

    def prepare_evaluate_hessian(
        self, inputs, hessian_inputs, adj_inputs, relevant_dependencies
    ):

        raise NotImplementedError
        tlm_inputs = [dep.tlm_value for dep in self.get_dependencies()]

        out_idx = []
        ipts = []
        for i, ip in enumerate(inputs):
            if ip is not None:
                out_idx.append(i)
                ipts.append(convert_arg(ip))

        N = len(self._vjpfuncs)
        hvp = [None] * N
        for (i, j), f in zip(
            product(self._tlm_argnums, self._tlm_argnums), self._hesfuncs
        ):
            x_dot = convert_arg(tlm_inputs[i])
            if x_dot is None:
                continue

            tmp = f(*ipts, *tuple(convert_arg(x) for x in adj_inputs), x_dot)
            if hvp[j] is None:
                hvp[j] = tmp
            else:
                hvp[j] = hvp[j] + tmp

        return hvp

    def evaluate_hessian_component(
        self,
        inputs,
        hessian_inputs,
        adj_inputs,
        block_variable,
        idx,
        relevant_dependencies,
        prepared=None,
    ):
        return prepared[idx]

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        out_idx = []
        ipts = []
        tips = []
        for i, (ip, ti) in enumerate(zip(inputs, tlm_inputs)):
            if ti is not None:
                out_idx.append(i)
                ipts.append(convert_arg(ip))
                tips.append(convert_arg(ti))
        jvp_maker = _pushforward(self._make_partial_diff_func(out_idx))
        jvpfun = jvp_maker(ipts)
        val = Enlist(jvpfun(*tips))
        return val

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]

    def prepare_evaluate_tlm_matrix(self, inputs, tlm_inputs, relevant_outputs):
        argnums = []
        for i, x in enumerate(tlm_inputs):
            if x is not None:
                argnums.append(i)
                for j, di_dj in enumerate(x):
                    if (i != j and di_dj is not None) or (
                        i == j and not isinstance(di_dj, JacobianIdentity)
                    ):
                        raise NotImplementedError(
                            "Non-identity inputs cannot be handled yet."
                        )
        outnums = [idx for idx, bv in relevant_outputs]

        recompute_jac = (tuple(argnums) != tuple(self._tlm_argnums)) or (
            tuple(outnums) != tuple(self._tlm_outnums)
        )
        if recompute_jac:
            self._tlm_argnums = argnums
            diff_func_relevant = self._get_reduced_output_diff_func(outnums)
            self._tlm_matrix_func = self._jit(
                jax.jacrev(diff_func_relevant, argnums=argnums)
            )

        cargs = tuple(convert_arg(x) for x in inputs)
        if not self._any_pointwise:
            val = self._tlm_matrix_func(*cargs)
            if self._single_output:
                return (val,)

            # Expand reduced outputs back to the expected length.
            rv = [None] * len(self.get_outputs())
            for i, idx in enumerate(outnums):
                rv[idx] = val[i]
            return rv

        # Set up all arguments into a standard format that can be looped through pointwise.
        outputs = [bv.saved_output for bv in self.get_outputs()]
        out_shapes = [
            a.shape[1:] if pw else a.shape
            for a, pw in zip(outputs, self._out_pointwise)
        ]
        n = max(
            [a.shape[0] for a, pw in zip(cargs, self._in_pointwise) if pw]
            + [a.shape[0] for a, pw in zip(outputs, self._out_pointwise) if pw]
        )

        standard_args = []
        for i, (a, pw) in enumerate(zip(cargs, self._in_pointwise)):
            if pw:
                if a.shape[0] not in (1, n):
                    raise ValueError(
                        f"Argument {i} is marked as pointwise but first axis doesn't match expected size ({a.shape[0]} != {n})"
                    )
                s = jnp.reshape(a, (a.shape[0], 1, *a.shape[1:]))
                s = jnp.broadcast_to(s, (n, *s.shape[1:]))
            else:
                s = jnp.broadcast_to(a, (n, *a.shape))
            standard_args.append(s)

        # v is a tuple of size n. Each one is a tuple of size num_relevant_outputs. Each entry in that tuple is a tuple of size num_relevant_inputs.

        doutput_dinput = [None] * len(self.get_outputs())
        squeeze_didj = jax.jit(partial(jnp.squeeze, axis=1))
        for i, out_idx in enumerate(outnums):
            out_pw = self._out_pointwise[out_idx]
            out_rank = len(out_shapes[out_idx])
            squeeze_didj_pw = jax.jit(partial(jnp.squeeze, axis=(1, 2 + out_rank)))
            squeeze_didj_in_pw = jax.jit(partial(jnp.squeeze, axis=1 + out_rank))
            if self._compiled_jacobian_stack and not self._compiled_jacobian:
                logger.warning(
                    "About to compile pointwise Jacobian-stacking function; this may take a few minutes..."
                )
                self._compiled_jacobian = True  # rather, it will be True in
                # a few lines once we call _pointjac_stack_XXX for the first
                # time
            di_dinput = [None] * len(argnums)
            for j, in_idx in enumerate(argnums):
                in_pw = self._in_pointwise[in_idx]

                if out_pw:
                    if self._single_output:
                        di_dj = _pointjac_stack_j(
                            self._tlm_matrix_func,
                            standard_args,
                            j,
                            self._compiled_jacobian_stack,
                        )
                    else:
                        di_dj = _pointjac_stack_ij(
                            self._tlm_matrix_func,
                            standard_args,
                            i,
                            j,
                            self._compiled_jacobian_stack,
                        )
                    if in_pw:
                        #  out_i: (n, *output_shape)
                        #  arg_j: (n, *input_shape)
                        #  di_dj: (n, 1, *output_shape, 1, *input_shape)
                        # di_dj': (n, *output_shape, *input_shape)
                        di_dj = squeeze_didj_pw(di_dj)
                    else:
                        #  out_i: (n, *output_shape)
                        #  arg_j: (*input_shape)
                        #  di_dj: (n, 1, *output_shape, *input_shape)
                        # di_dj': (n, *output_shape, *input_shape)
                        di_dj = squeeze_didj(di_dj)  # _dj)
                elif in_pw:
                    #  out_i: (*output_shape)
                    #  arg_j: (n, *input_shape)
                    #  di_dj: (*output_shape, n, 1, *input_shape)
                    # di_dj': (*output_shape, n, *input_shape)
                    di_dj = _pointjac_stack_ij_out(
                        self._tlm_matrix_func,
                        standard_args,
                        i,
                        j,
                        out_rank,
                        self._compiled_jacobian_stack,
                    )
                    di_dj = squeeze_didj_in_pw(di_dj)
                else:
                    # If neither is defined pointwise, the shapes should be this:
                    #  out_i: (*output_shape)
                    #  arg_j: (*input_shape)
                    #  di_dj: (n, *output_shape, *input_shape)
                    # di_dj': (*output_shape, *input_shape)
                    di_dj = _pointjac_stack_ij(
                        self._tlm_matrix_func,
                        standard_args,
                        i,
                        j,
                        self._compiled_jacobian_stack,
                    )
                    di_dj = di_dj[0, ...]
                di_dinput[j] = di_dj
            doutput_dinput[out_idx] = di_dinput
        return doutput_dinput

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        relevant_outputs_idxs = [
            idx for idx, adj in enumerate(adj_inputs) if adj is not None
        ]
        diff_func_relevant = self._get_reduced_output_diff_func(relevant_outputs_idxs)

        outs, fvjp = jax.vjp(diff_func_relevant, *tuple(convert_arg(x) for x in inputs))
        fvjp = self._jit(fvjp)

        adj_inputs = tuple(
            convert_arg(adj_inputs[idx]) for idx in relevant_outputs_idxs
        )
        if self._single_output:
            adj_inputs = adj_inputs[0]
        return Enlist(fvjp(adj_inputs))

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]


def _pointjac_stack_j(jacfun, standard_args, j, compiled=True):
    if compiled:
        return jax.jit(jnp.stack)([jacfun(*args)[j] for args in zip(*standard_args)])
    else:
        return jnp.stack([jacfun(*args)[j] for args in zip(*standard_args)])


def _pointjac_stack_ij(jacfun, standard_args, i, j, compiled=True):
    if compiled:
        return jax.jit(jnp.stack)([jacfun(*args)[i][j] for args in zip(*standard_args)])
    else:
        return jnp.stack([jacfun(*args)[i][j] for args in zip(*standard_args)])


def _pointjac_stack_ij_out(jacfun, standard_args, i, j, out_ax, compiled=True):
    if compiled:
        return jax.jit(partial(jnp.stack, axis=out_ax))(
            [jacfun(*args)[i][j] for args in zip(*standard_args)]
        )
    else:
        return partial(jnp.stack, axis=out_ax)(
            [jacfun(*args)[i][j] for args in zip(*standard_args)]
        )


def _pushforward(func):
    def prepare_jvp(inputs):
        outputs, jvp_func = jax.linearize(func, *inputs)
        return jax.jit(jvp_func)

    return prepare_jvp


def _pullback(func, argnum):
    def prepare_vjp(inputs, adj_inputs):
        outputs, vjp_func = jax.vjp(func, *inputs)
        vjp_func = jax.jit(vjp_func)
        return vjp_func(*adj_inputs)

    def eval_vjp_component(prepared):
        return prepared[argnum]

    return prepare_vjp, eval_vjp_component


def _make_hvpfun(f):
    def make_hvp(inputs, adj_inputs):
        outs, fvjp = jax.vjp(f, *tuple(convert_arg(x) for x in inputs))
        outs, fhvp = jax.linearize(
            jax.jit(fvjp), *tuple(convert_arg(x) for x in adj_inputs)
        )
        return jax.jit(fhvp)

    return make_hvp


def vector_jacobian_product(fun, argnum, reverse=True, jitfun=jax.jit):
    # based on the old autograd implementation
    def vec_prod(*args, **kwargs):
        args, x_dot = args[:-1], args[-1]
        return jnp.tensordot(x_dot, fun(*args, **kwargs), axes=jnp.ndim(x_dot))

    if reverse:
        return jax.jacrev(vec_prod, argnums=argnum)
    return jitfun(jax.jacfwd(vec_prod, argnums=argnum))


def is_tracer(x):
    return isinstance(x, jax.core.Tracer)


def get_overloaded(x: Any) -> OverloadedType:
    if isinstance(x, (ndarray, tuple)):
        return x
    return create_overloaded_object(x)


def overload_jax(
    func: Function,
    static_argnums: Optional[Union[int, Iterable[int]]] = None,
    argnums: Optional[Union[int, Iterable[int]]] = None,
    nojit: Optional[bool] = False,
    function_name: Optional[str] = None,
    checkpoint: bool = False,
    concrete: bool = False,
    backend: Optional[str] = None,
    donate_argnums: Optional[Union[int, Iterable[int]]] = None,
    pointwise: Union[bool, Sequence[bool]] = False,
    out_pointwise: Optional[Union[bool, Sequence[bool]]] = None,
    compile_jacobian_stack: bool = True,
    **jax_kwargs,
) -> OverloadedFunction:
    """
    Creates a pyadjoint-overloaded version of a JAX-traceable function.

    :param func: The function to JIT compile and make differentiable
    :type func: Function
    :param static_argnums: The static_argnums parameter of jax.jit (e.g. numbers
        of arguments that, if changed, should trigger recompilation)
    :type static_argnums: Union[int,Iterable[int]], optional
    :param argnums: The numbers of the arguments you want to differentiate
        with respect to. For example, if you have a function f(x,p,w) and
        want the derivative with respect to p and w, pass argnums=(1,2).
    :type argnums: Union[int,Iterable[int]], optional
    :param nojit:  If True, do NOT JIT compile the function, defaults to False
    :type nojit: bool, optional
    :param function_name:  if you want the function's name on the JAXBlock recorded as
        something other than func.__name__, use this parameter
    :type function_name: str, optional
    :param checkpoint: if True, make `func` recompute internal linearization
        points when differentiated (as opposed to computing these in the forward
        pass and storing the results). This increases total FLOPs in exchange for
        less memory usage/fewer acceses, defaults to False
    :type checkpoint: bool, optional
    :param concrete: if True, indicates that the function requires value-dependent
        Python control flow, defaults to False
    :type concrete: bool, optional
    :param backend: String representing the XLA backend to use (e.g. 'cpu', 'gpu',
           'tpu'). (Note that this is an experimental JAX feature, and its API is
           likely to change), defaults to None
    :type backend: str, optional
    :param donate_argnums: Which arguments are 'donated' to the computation? In other
           words, you cannot reuse these arguments after calling the function. This
           lets XLA more aggresively re-use donated buffers., default None
    :type donate_argnums: Union[int,Iterable[int]], optional
    :param pointwise: By default, this is false. True means the function performs
        operations on a batch of points. This allows optimizing the Jacobian calculations
        by only computing the diagonal component. If a list, then there should be a bool
        for each argnum.
    :type pointwise: Union[bool,Sequence[bool]], optional
    :param out_pointwise: If any inputs are defined pointwise, this specifies which
        outputs are defined pointwise. If a list, then there should be abool for each
        output. By default, all outputs will be assumed pointwise if any inputs are
        pointwise.
    :type out_pointwise: Union[bool,Sequence[bool]], optional
    :param compile_jacobian_stack: If we need to make a Jacobian matrix
        corresponding to this function, and that computation is done pointwise,
        we have to `jnp.stack` the pointwise Jacobians into one full Jacobian.
        This parameter controls whether or not we compile that stack operation
        with `jax.jit`, which can take several minutes to compile the stack
        the first time a Jacobian is requested (since it requires unrolling
        a Python-mode loop over quadrature points), but often leads to
        factor-of-5-or-more improvements in the total time
        required to compute the Jacobian for subsequent evaluations. Defaults to True
    :type compile_jacobian_stack: bool, optional

    :return: A function with the same signature as `func` that performs the same
      computation, just with JAX JIT compilation, and being differentiable
      by both JAX and Pyadjoint
    :rtype: OverloadedFunction
    """

    if checkpoint:
        func = jax.checkpoint(func, concrete=concrete)
    # first, jit compile
    if not nojit:
        static_argnums = static_argnums or ()
        donate_argnums = donate_argnums or ()
        jit_kwargs = dict(
            static_argnums=static_argnums,
            backend=backend,
            donate_argnums=donate_argnums,
            **jax_kwargs,
        )
        func = jax.jit(func, **jit_kwargs)
    else:
        jit_kwargs = None
    # now overload

    @wraps(func)
    def _overloaded_func(*args, **kwargs):
        annotate = annotate_tape(kwargs)
        with stop_annotating():
            cargs = tuple(convert_arg(x) for x in args)
            out = func(*cargs, **kwargs)
            if is_tracer(out):
                return out
            for arg in cargs:
                if is_tracer(arg):
                    return out

        out = Enlist(out)
        overloads = [create_overloaded_object(arr) for arr in out]
        if annotate:
            tape = get_working_tape()
            kwargs["func_name"] = function_name
            block = JAXBlock(
                func,
                args,
                overloads,
                argnums=argnums,
                nojit=nojit,
                jit_kwargs=jit_kwargs,
                pointwise=pointwise,
                out_pointwise=out_pointwise,
                compile_jacobian_stack=compile_jacobian_stack,
                **kwargs,
            )
            tape.add_block(block)

        return out.delist(overloads)

    return _overloaded_func



from setuptools import setup, find_packages
from itertools import chain
import os
import pathlib

python_requires = ">=3.7"
install_requires = [
    "dolfin-adjoint >= 2019.1.1",
    "numpy",
    "autograd",
    "petsc4py>=3.11.0",
    "jax",
    "jaxlib",
]
extras_requires = {
    "test": ["pytest>=3.10", "pytest-rerunfailures", "flake8", "coverage"],
    "visualization": ["matplotlib"],
    "flax": ["flax"],
    "tensorflow": ["tensorflow>=2.0"],
    "doc": [
        "sphinx",
        "myst-nb",
        "sphinxcontrib-bibtex",
        "sphinxcontrib-katex",
        "pydata-sphinx-theme",
        "linkify-it-py",
    ],
    "dev": ["black", "pre-commit", "build"],
}

extras_requires["all"] = list(chain(*extras_requires.values()))

dependency_links = [
    "git+https://github.com/dolfin-adjoint/pyadjoint.git@master",
]

classifiers = [
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]

url = "https://gitlab.com/crikit/crikit"

project_urls = {
    "Bug Tracker": "https://gitlab.com/crikit/crikit/issues",
    "Documentation": "https://crikit.science/documentation",
}

version_file = os.path.join(os.path.dirname(__file__), "crikit/_version.py")
with open(version_file, "r") as f:
    version = f.read().split("=", 1)[1].strip(" \n\"'")


HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="crikit",
    version=version,
    description="Constitutive Relation Inference Toolkit",
    long_description=README,
    long_description_content_type="text/markdown",
    author="CRIKit Team",
    classifiers=classifiers,
    url=url,
    project_urls=project_urls,
    packages=find_packages(exclude=["examples", "tests"]),
    python_requires=python_requires,
    install_requires=install_requires,
    dependency_links=dependency_links,
    extras_require=extras_requires,
)



from .types import Space, PointMap
from .space_builders import DirectSum
import numpy as np
from jax import numpy as jnp
from jax.interpreters.xla import DeviceArray


class Ndarrays(Space):
    """This class represents a Space of NumPy arrays of a given shape and
    optionally a specific data type.

    Negative numbers can be used in the shape to indicate that the length of
    that dimension doesn't matter as long as that dimension exists. For example,
    if the given shape is (-1, 2, 5), then arrays with shapes (4, 2, 5) and (1, 2, 5)
    are both points in the space, but (2, 5) is not.

        >>> import numpy as np
        >>> from crikit.cr.numpy import Ndarrays
        >>> space = Ndarrays((-1, 2, 5))
        >>> space.is_point(np.zeros((4, 2, 5)))
        True
        >>> space.is_point(np.zeros((1, 2, 5)))
        True
        >>> space.is_point(np.zeros((2, 5)))
        False

    Args:
        shape (tuple or list): The shape of the arrays.
        dtype (numpy.dtype): A NumPy datatype that further constrains the Space.
            If None, Ndarrays space will not have a specific type.
    """

    def __init__(self, shape, dtype=None):
        self._shape = shape
        self._dtype = dtype
        self._indefinite_axes = np.array(shape) < 0
        self._definite_axes = np.logical_not(self._indefinite_axes)
        self._definite_shape = np.asarray(shape)[self._definite_axes]

    def shape(self):
        return self._shape

    def is_point(self, point):
        """Returns true if the given point is an ndarray and its shape and dtype
        match those of the space.
        """
        if not isinstance(point, (np.ndarray, jnp.ndarray, DeviceArray)):
            return False
        if len(point.shape) != len(self._shape):
            return False
        # Make sure that the shapes are the same along each definite axis
        point_definite_shape = np.asarray(point.shape)[self._definite_axes]
        return np.array_equal(point_definite_shape, self._definite_shape) and (
            self._dtype is None or point.dtype == self._dtype
        )

    def point(self, **kwargs):
        shape = self._shape
        if "near" in kwargs:
            near = kwargs["near"]
            assert self.is_point(near)
            shape = near.shape
        return np.zeros(shape)

    def __eq__(self, other):
        return (
            isinstance(other, Ndarrays)
            and self._shape == other._shape
            and self._dtype == other._dtype
        )

    def __repr__(self):
        if self._dtype is None:
            return f"Ndarrays({self._shape})"
        return f"Ndarrays({self._shape}, dtype={self._dtype})"


class CR_P_LaplacianNumpy(PointMap):
    def __init__(self, p=2, dim=2, input_u=True):
        self._p = p
        self._input_u = input_u
        np_vec_space = Ndarrays((-1, dim))

        if self._input_u:
            np_scalar_space = Ndarrays((-1,))
            source = DirectSum(np_scalar_space, np_vec_space)
        else:
            source = np_vec_space
        super(CR_P_LaplacianNumpy, self).__init__(source, np_vec_space)

    def __call__(self, args, **kwargs):
        if self._input_u:
            gradu = args[1]
        else:
            gradu = args

        mu = (np.sum(gradu * gradu, axis=1) + 1e-12) ** ((self._p - 2) / 2)
        mu = mu[:, None]
        out = mu * gradu

        return out

    def setParams(self, p):
        self._p = p



from pyadjoint.enlisting import Enlist
from pyadjoint.tape import stop_annotating, get_working_tape

from .drivers import (
    compute_gradient,
    compute_jacobian_action,
    compute_jacobian_matrix,
    compute_hessian_action,
)

# This is copied and extended from work by Sebastian Mitusch in pyadjoint
# [https://bitbucket.org/dolfin-adjoint/pyadjoint/commits/4b67f3e07579501ab95b1fad143d98253a15f7ae?at=reduced-function]


class ReducedFunction(object):
    def __init__(
        self,
        outputs,
        controls,
        tape=None,
        eval_cb_pre=None,
        eval_cb_post=None,
        jac_action_cb_pre=None,
        jac_action_cb_post=None,
        adj_jac_action_cb_pre=None,
        adj_jac_action_cb_post=None,
        hess_action_cb_pre=None,
        hess_action_cb_post=None,
    ):
        self.output_vals = Enlist(outputs)
        self.outputs = Enlist([out.block_variable for out in self.output_vals])
        self.outputs.listed = self.output_vals.listed

        self.controls = Enlist(controls)
        self.tape = get_working_tape() if tape is None else tape

        nothing = lambda *args: None
        self.eval_cb_pre = nothing if eval_cb_pre is None else eval_cb_pre
        self.eval_cb_post = nothing if eval_cb_post is None else eval_cb_post
        self.jac_action_cb_pre = (
            nothing if jac_action_cb_pre is None else jac_action_cb_pre
        )
        self.jac_action_cb_post = (
            nothing if jac_action_cb_post is None else jac_action_cb_post
        )
        self.adj_jac_action_cb_pre = (
            nothing if adj_jac_action_cb_pre is None else adj_jac_action_cb_pre
        )
        self.adj_jac_action_cb_post = (
            nothing if adj_jac_action_cb_post is None else adj_jac_action_cb_post
        )
        self.hess_action_cb_pre = (
            nothing if hess_action_cb_pre is None else hess_action_cb_pre
        )
        self.hess_action_cb_post = (
            nothing if hess_action_cb_post is None else hess_action_cb_post
        )

    def jac_action(self, inputs, options=None):
        inputs = Enlist(inputs)
        if len(inputs) != len(self.controls):
            raise TypeError(
                "The length of inputs must match the length of function controls."
            )

        values = [c.data() for c in self.controls]
        self.jac_action_cb_pre(
            self.controls.delist(values), self.controls.delist(inputs)
        )

        derivatives = compute_jacobian_action(
            self.output_vals, self.controls, inputs, options=options, tape=self.tape
        )
        # Call callback
        self.jac_action_cb_post(
            self.outputs.delist([bv.saved_output for bv in self.outputs]),
            self.outputs.delist(derivatives),
            self.controls.delist(values),
        )

        return self.outputs.delist(derivatives)

    def adj_jac_action(self, inputs, options=None):
        inputs = Enlist(inputs)
        if len(inputs) != len(self.outputs):
            raise TypeError(
                "The length of inputs must match the length of function outputs."
            )

        values = [c.data() for c in self.controls]
        self.adj_jac_action_cb_pre(self.controls.delist(values))

        derivatives = compute_gradient(
            self.output_vals,
            self.controls,
            options=options,
            tape=self.tape,
            adj_value=inputs,
        )

        # Call callback
        self.adj_jac_action_cb_post(
            self.outputs.delist([bv.saved_output for bv in self.outputs]),
            self.controls.delist(derivatives),
            self.controls.delist(values),
        )

        return self.controls.delist(derivatives)

    def jac_matrix(self, m_jac=None):
        if m_jac is not None:
            m_jac = Enlist(m_jac)
            if len(m_jac) != len(self.controls):
                raise TypeError(
                    "The length of m_jac must match the length of function controls."
                )

            for i, jac in enumerate(m_jac):
                m_jac[i] = Enlist(jac)
                if len(m_jac[i]) != len(self.controls):
                    raise TypeError(
                        "The length of each identity must match the length of function controls."
                    )

        jacobian = compute_jacobian_matrix(
            self.output_vals, self.controls, m_jac, tape=self.tape
        )
        for i, jac in enumerate(jacobian):
            if jac is not None:
                jacobian[i] = self.controls.delist(jac)
        jacobian = self.outputs.delist(jacobian)
        return jacobian

    def hess_action(self, m_dot, adj_input, options=None):
        m_dot = Enlist(m_dot)
        if len(m_dot) != len(self.controls):
            raise TypeError(
                "The length of m_dot must match the length of function controls."
            )

        adj_input = Enlist(adj_input)
        if len(adj_input) != len(self.outputs):
            raise TypeError(
                "The length of adj_input must match the length of function outputs."
            )

        values = [c.data() for c in self.controls]
        self.hess_action_cb_pre(self.controls.delist(values))

        derivatives = compute_gradient(
            self.output_vals,
            self.controls,
            options=options,
            tape=self.tape,
            adj_value=adj_input,
        )

        # TODO: there should be a better way of generating hessian_input.
        zero = [0 * v for v in adj_input]
        hessian = compute_hessian_action(
            self.output_vals,
            self.controls,
            m_dot,
            options=options,
            tape=self.tape,
            hessian_value=zero,
        )

        # Call callback
        self.hess_action_cb_post(
            self.outputs.delist([bv.saved_output for bv in self.outputs]),
            self.controls.delist(hessian),
            self.controls.delist(values),
        )

        return self.controls.delist(hessian)

    def __call__(self, inputs):
        inputs = Enlist(inputs)
        if len(inputs) != len(self.controls):
            raise TypeError("The length of inputs must match the length of controls.")

        # Call callback.
        self.eval_cb_pre(self.controls.delist(inputs))

        for i, value in enumerate(inputs):
            self.controls[i].update(value)

        # self.tape.reset_blocks()
        with self.marked_controls():
            with stop_annotating():
                self.tape.recompute()

        output_vals = self.outputs.delist(
            [output.checkpoint for output in self.outputs]
        )

        # Call callback
        self.eval_cb_post(output_vals, self.controls.delist(inputs))

        return output_vals

    def marked_controls(self):
        return marked_controls(self)


class marked_controls(object):
    def __init__(self, rf):
        self.rf = rf

    def __enter__(self):
        for control in self.rf.controls:
            control.mark_as_control()

    def __exit__(self, *args):
        for control in self.rf.controls:
            control.unmark_as_control()



from .cr import CR, BlockCR, cr_function_shape, save_jax_cr, RivlinModel, P_Laplacian
from .jax_utils import (
    JAXArrays,
    JAX_To_UFLFunctionSpace,
    UFLExprSpace_To_JAX,
    ReducedFunctionJAX,
    JAX_UFLFunctionSpace_Covering,
)
from .ufl import (
    UFLFunctionSpace,
    UFLExprSpace,
    CR_UFL_Expr,
    create_ufl_standins,
    create_ufl_standin_arguments,
)
from .numpy import Ndarrays
from .fe import (
    form_get_expr_space,
    assemble_with_cr,
    get_ufl_composite_cr,
    get_cr_form_degree,
    AssembleWithCRBlock,
)
from .types import Space, PointMap
from .map_builders import (
    Callable,
    AugmentPointMap,
    Parametric,
    CompositePointMap,
    ParallelPointMap,
    IdentityPointMap,
)
from .space_builders import DirectSum



from crikit.cr.types import Space
from crikit.cr.space_builders import DirectSum
from crikit.cr.map_builders import Callable
from functools import update_wrapper


class Numbers(Space):
    def __init__(self, name, pointclass, compat_classes=()):
        self._name = name
        self._pointclass = pointclass
        self._compat_classes = frozenset(compat_classes + (pointclass,))

    def shape(self):
        return ()

    def __str__(self):
        return self._name

    def point(self, **kwargs):
        return self._pointclass(0)

    def is_point(self, point):
        if isinstance(point, self._compat_classes):
            return True
        return False


class Integers(Numbers):
    def __init__(self, name=None):
        _name = "ZZ(std)" if name is None else name
        super(Integers, self).__init__(_name, int)


ZZ = Integers()  #: Represents the space of integers


class Reals(Numbers):
    def __init__(self, name=None):
        _name = "RR(std)" if name is None else name
        super(Reals, self).__init__(_name, float, (int,))


RR = Reals()  #: Represents the space of real numbers


class Complexs(Numbers):
    def __init__(self, name=None):
        _name = "CC(std)" if name is None else name
        super(Complexs, self).__init__(_name, complex, (int, float))


CC = Complexs()  #: Represents the space of complex numbers


def type_tuple_to_space(tt):
    """Converts the given numeric type or tuple of types to a Space using the
    stdnumeric spaces and DirectSum.

    If ``tt`` is a tuple, :code:`type_tuple_to_space()` is recursively called on
    each element of the tuple, and the DirectSum of the result is returned.

    If ``tt`` is a numeric type, then the corresponding stdnumeric Space is
    returned. If ``tt`` is a subclass of a numeric type, the name of the
    returned space will contain the name of that class.

    Supported numeric types are :class:`int`, :class:`float`, :class:`complex`,
    or subclasses of those types.

    Args:
        tt: a numeric type or arbitrarily nested tuples of numeric types.

    Returns:
        Space: the stdnumeric Space corresponding to the given types.

    Todo:
        * Use numbers.Integral/Real/Complex to find the corresponding space,
    """
    if isinstance(tt, type):
        stdnumeric = (
            (int, ZZ, Integers, "ZZ"),
            (float, RR, Reals, "RR"),
            (complex, CC, Complexs, "CC"),
        )
        for numbers in stdnumeric:
            if issubclass(tt, numbers[0]):
                if tt == numbers[0]:
                    return numbers[1]
                else:
                    return numbers[2](f"{numbers[3]}({tt!s})")
        raise NotImplementedError
    elif isinstance(tt, tuple):
        return DirectSum(*tuple(map(type_tuple_to_space, tt)))
    else:
        raise NotImplementedError


def point_map(source_types, target_types, **kwargs):
    """Decorates a function to make it a point map (by constructing a Callable
    instance).

    Args:
        source_types: a type tuple representing the source space.
        target_types: a type tuple representing the target space.
        **kwargs: passed through to the :class:`~crikit.cr.map_builders.Callable` constructor.

    Here's example usage for calculating the p-norm of a two-dimensional vector:

    .. testcode::

        from crikit.cr.stdnumeric import point_map

        @point_map(((float, float), float), float, bare=True)
        def pnorm_2d(v, p):
            return (v[0]**p + v[1]**p) ** (1/p)
        assert pnorm_2d(((1, 2), 1)) == 3

        from crikit.cr.types import PointMap
        from crikit.cr.space_builders import DirectSum
        from crikit.cr.stdnumeric import RR

        assert isinstance(pnorm_2d, PointMap)
        assert pnorm_2d.source == DirectSum(DirectSum(RR, RR), RR)
        assert pnorm_2d.target == RR
    """

    source_space = type_tuple_to_space(source_types)
    target_space = type_tuple_to_space(target_types)

    def point_map_decorator(func):
        pm = Callable(source_space, target_space, func, **kwargs)
        return pm

    return point_map_decorator



from pyadjoint.tape import get_working_tape, annotate_tape
from pyadjoint_utils.block import Block
from pyadjoint_utils.tape import push_tape
from contextlib import contextmanager


@contextmanager
def record_tape_block(**kwargs):
    with push_tape() as new_tape:
        yield new_tape
    add_tape_block(new_tape, **kwargs)


def add_tape_block(tape, **kwargs):
    annotate = annotate_tape(kwargs)
    if annotate:
        working_tape = get_working_tape()
        if tape == working_tape:
            raise ValueError("TapeBlock must be recorded on a separate tape")
        block = TapeBlock(tape, **kwargs)
        # TODO: should I have a check here to make sure the same tape bock isn't added twice?
        working_tape.add_block(block)


class TapeBlock(Block):
    """A block that represents an entire tape.

    This idea could also be expressed as a ReducedFunctionBlock. In many cases,
    the user wll already know what the desired dependencies and outputs are. If
    they give the ReducedFunction, then it is possible to avoid computing blocks
    that don't show up in the list of the ReducedFunction's outputs.
    """

    def __init__(self, tape, name=None):
        super().__init__()
        self.tape = tape

        # Note: add_dependency() and add_output() are not used here because they
        #   call the _ad_will_add_as_xxxx() functions, which sometimes make
        #   copies of the variable. The tape visualization (and maybe
        #   computation?) will be messed up if there are extraneous
        #   BlockVariable copies.
        dependencies, outputs = self.tape.find_absolute_dependencies_outputs()
        self._dependencies += dependencies
        self._outputs += outputs

        self.name = name if name is not None else "TapeBlock"

    def __str__(self):
        return self.name

    @property
    def tf_name(self):
        return self.name

    def reset_variables(self, types=None):
        self.tape.reset_variables(types)

    def prepare_recompute_component(self, inputs, relevant_outputs):
        block_variables = [bv for i, bv in relevant_outputs]
        self.tape.recompute(outputs=block_variables)

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        block_variables = [bv for i, bv in relevant_dependencies]
        # FIXME: I actually have to overwrite evaluate_adj() to get the correct
        # value for markings, but it's usually going to be True.
        self.tape.evaluate_adj(inputs=block_variables, markings=True)

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        block_variables = [bv for i, bv in relevant_outputs]
        self.tape.evaluate_tlm(outputs=block_variables)

    def prepare_evaluate_hessian(
        self, inputs, hessian_inputs, adj_inputs, relevant_dependencies
    ):
        block_variables = [bv for i, bv in relevant_dependencies]
        self.tape.evaluate_hessian(inputs=block_variables, markings=True)

    def prepare_evaluate_tlm_matrix(self, inputs, tlm_inputs, relevant_outputs):
        block_variables = [bv for i, bv in relevant_outputs]
        self.tape.evaluate_tlm_matrix(outputs=block_variables, markings=True)

    # These functions don't need to do anything because everything is handled in the 'prepare_*' functions.
    def _nothing(*args, **kwargs):
        pass

    recompute_component = _nothing
    evaluate_adj_component = _nothing
    evaluate_tlm_component = _nothing
    evaluate_tlm_matrix_component = _nothing
    evaluate_hessian_component = _nothing

    def tf_get_blocks(self):
        return (self, self.tape.get_blocks())
